import inspect
import os
import abc
import time
import sys
import logging
import platform
import traceback

from quarchpy.disk_test.Drive_wrapper import DriveWrapper
from quarchpy.disk_test.hostInformation import HostInformation
from quarchpy.disk_test.driveTestCore import checkDriveState, is_tool, get_module_from_choice, get_quarch_modules_qps
from quarchpy.disk_test.dtsComms import DTSCommms
from quarchpy.disk_test.dtsGlobals import dtsGlobals
from quarchpy.disk_test.Custom_test_variable import CustomVariable
from quarchpy.qps.qpsFuncs import isQpsRunning
from quarchpy.device.scanDevices import scanDevices
import inspect

ABC = abc.ABCMeta('ABC', (object,), {})


# This could use a comment
class IdGenerator:
    def __init__(self, base_test_point):
        # Why not track number of tiers as len(unique_ID)?

        # Array of "tiers" for ID's
        self.unique_ID = [0]
        # what 'tier' we are one / changing
        self.tier_level = 0

        self.first_call = True

        self.test_ids = []
        # tier_levels : 0   1   2   3
        # tier_Values : 0 , 0 , 0 , 0
        self.base = base_test_point

        if self.base is None:
            self.base = print

    def reset(self):
        self.unique_ID = [0]
        self.tier_level = 0
        self.first_call = True

    def up_tier(self, description=None, singular=False, add_to_current_tier=False):

        if add_to_current_tier:
            self.base(self.gen_next_id(), test_description=description)

        # Auto add another tier level if required
        if self.tier_level is (len(self.unique_ID) - 1):
            self.unique_ID.append(0)
        self.tier_level += 1

        if description and not add_to_current_tier:
            self.base(self.gen_next_id(), test_description=description)

        if not singular:
            if self.tier_level is (len(self.unique_ID) - 1):
                self.unique_ID.append(0)
            self.tier_level += 1

    def down_tier(self, singular=False, description=None):
        # Cannot decrease more than 0'th tier
        if self.tier_level == 0:
            return
        self.tier_level -= 1

        if description:
            self.base(self.gen_next_id(), test_description=description)

        if not singular:
            self.tier_level -= 1

    def return_current(self):
        strings = [str(tier_value) for tier_value in self.unique_ID]
        unique_id = ".".join(strings)

        return str(unique_id)

    def return_parent_id(self):
        strings = [str(tier_value) for tier_value in self.unique_ID]
        unique_id = ".".join(strings)

        # Return itself if no parent
        if "." not in unique_id:
            return str(unique_id)

        # remove last tier value
        unique_id = unique_id[:unique_id.rfind(".")]

        return str(unique_id)

    def gen_next_id(self):

        for i in range(len(self.unique_ID)):
            if i is self.tier_level:
                if self.first_call:
                    self.first_call = False
                    break
                self.unique_ID[self.tier_level] += 1

        # reset all additional tiers to 0
        while (len(self.unique_ID) - 1) > self.tier_level:
            # Remove last index
            self.unique_ID.pop()

        strings = [str(tier_value) for tier_value in self.unique_ID]
        unique_id = ".".join(strings)

        self.test_ids.append(unique_id)
        return str(unique_id)


def _create_scan_dict(use_qps, ip_lookup, filter_module):
    """
    :return: A dictionary of arguments appropriate for module search function
    """
    if not use_qps:
        return {"ipAddressLookup": ip_lookup, "module_type_filter": filter_module}
    else:
        return {}


class BaseTest(ABC):
    def __init__(self):
        # pass base_test, so ID generator can add test points.
        self.test_id = IdGenerator(self.test_point)
        self.comms = DTSCommms()
        self.my_host_info = HostInformation()

        self.number_of_test_points = 0
        self.report = {}

        self.custom_variable_dict = {}
        self.skipped_tests = []
        self.user_vars = []
        self.doc_counter = 0
        self.test_points = {}

        self.test_errors = []

        self.document_mode = False
        self.execution_mode = "run"

    def _set_documentation_mode(self, document_mode=False):
        if "seek" in str(document_mode):
            # Mode to seek number of test points
            self.execution_mode = "seek"
            self.number_of_test_points = 0
            self.document_mode = True
            return
        if document_mode:
            self.execution_mode = "document"
            self.document_mode = True
        else:
            self.execution_mode = "run"
            self.document_mode = False

    def ask_for_user_vars(self):
        self.custom_variable_dict = self._get_custom_variables()
        self.skipped_tests = self._list_test_to_skip()
        self.__change_custom_vars()

    def __change_custom_vars(self):
        if not self.custom_variable_dict:
            return

        # Loop through each custom variable - It's a list of objects so should be mutable
        for variable in self.user_vars:
            # Check if variable name in return dictionary
            if variable.custom_name in self.custom_variable_dict.keys():
                # replace custom variable value with new value
                variable.custom_value = self.custom_variable_dict[variable.custom_name]

    def declare_custom_variable(self, default_value, custom_name, description="", var_purpose=None, accepted_vals=None,
                                numerical_max=None):
        """

        :param default_value: Default value of custom variable
        :param custom_name: Custom value is added by user in java GUI
        :param description: Short description of variable use in test
        :param var_purpose: > USED LIKE AN ENUM > Tells us if custom variable is user editable
        :param accepted_vals: List of accepted values for item ( Shown as dropdown in GUI )
        :param numerical_max: Allows a maximum value to be applied to variable (MINIMUM VALUE ALWAYS 0)
        :return:
        """

        new_var = CustomVariable(name=custom_name, description=description, default_value=default_value,
                                 var_purpose=var_purpose, accepted_vals=accepted_vals, numerical_max=numerical_max)

        # Add to user variables list
        if not var_purpose:
            self.user_vars.append(new_var)

        # return a new custom variable instance
        return new_var

    def _list_test_to_skip(self):
        """
        Tests to skip included inside custom variables

        Format:
        1-3 --> Skip all tests points beginning with 1,2 or 3
        1.3 --> Skip test point 1.3
        1.4,1.5 --> Skip test points 1.4 and 1.5

        :return: List of all ID's / tests to be skipped
        """
        if "skip_tests" in self.custom_variable_dict.keys():
            # return list of id's to skip
            return self._return_skipable_tests(self.custom_variable_dict["skip_tests"])
        else:
            return []  # Return an empty list

    def _return_skipable_tests(self, string_list):
        # split them at a comma
        string_divide = string_list.split(",")

        items_to_skip = []

        for item in string_divide:
            if "-" in item:
                # should only ever be in form '1-4'
                range_split = item.split("-")
                # Consider smaller form below. Also does range_split need to be forced to int?
                # This could probably also use a try/catch statement in case user enters an
                # incorrect range
                temp_val = []
                items_to_skip += range(int(range_split[0]), int(range_split[1]) + 1)
                for new_item in temp_val:
                    items_to_skip.append(str(new_item))

            else:
                # if item isn't a range, it's just a test point to skip > '4.7
                items_to_skip.append(str(item).strip())

        # converting everything to string
        for index, item in enumerate(items_to_skip):
            if isinstance(item, int):
                items_to_skip[index] = str(item)

        return items_to_skip

    def _skip_test(self, unique_id):
        """
        :param unique_id: ID to check against skipped tests
        :return: where or not to skip the test point
        """
        for item in self.skipped_tests:
            if str(unique_id[:1]) == str(item):
                return True
            else:
                if unique_id == item:
                    return True
        return False

    def start_test(self, document_mode=False):
        """
            Base function for starting a test.
            All tests passed from QCS are required to override this function
        """
        pass

    def check_prerequisites(self):
        """
            Base function for checking any import modules of a test
            All tests require LSPCI / WMIC for windows
            All tests require LSPCI / LSSCSI for linux
        """
        # if not on windows, check lspci and lsscsi are installed
        if os.name != 'nt':
            if not is_tool("lsscsi"):
                self.test_errors.append("Lsscsi not found on server machine")
            if not is_tool("lspci"):
                self.test_errors.append("Lspci not found on server machine")

    def _get_custom_variables(self):
        """
            Base function, once a test class is instantiated, this method is called
            All tests passed from QCS will call this in order to get dictionary of custom variables
        """

        return self.comms.sendMsgToGUIwithResponse(self.comms.create_request_variable(self.user_vars),
                                                   timeToWait=None)

    def gather_initial_report_values(self):
        self.report.clear()

        # add custom variables in test
        self.report["custom_variables"] = self.custom_variable_dict

        # add hardware information
        self.report["cpu"] = platform.processor()
        self.report["operating_system"] = platform.system() + " " + platform.version()
        self.report["host_platform"] = platform.node()

    def seek_test_values(self):
        # Run through test, counting up all the test / check points
        try:
            self.start_test(document_mode="seek")
        except Exception as e:
            self.test_errors.append("Error during seek of test : " + str(e))
        # send a message to say test consists of x amount of test points
        self.comms.sendMsgToGUI(self.comms.create_request_status(number_of_test_points=self.number_of_test_points))
        # gui will auto increment counter / progress with every log sent over

    def test_point(self, unique_id=None, function=None, function_args=None, function_description=None, debug=None,
                   warning=None, test_description=None, has_return=False, stop_timeout=False):

        """
        :param stop_timeout: Indicates whether function needs to stop QCS timeout
        :param test_description: Relays only a test description to QCS
        :param debug: Relays only debug information to QCS
        :param has_return: Whether to return item from function
        :param unique_id: test point ID
        :param function_description: describes what function to be executed is doing
        :param function: name of function to execute
        :param function_args: Arguments for function being executed

        :return: pass / fail
        """

        # Skipping tests, adding to tally - Used primarily in progress bar.
        if self.execution_mode == "seek":
            self.number_of_test_points += 1

            src_string = inspect.getsource(sys._getframe().f_back)
            print(src_string.count("test_point"))

            self.number_of_test_points += src_string.count("test_point")

            if has_return:
                return "Unused String"
            return None

        if not dtsGlobals.continueTest:
            return

        return_obj = ReturnObject()

        # checking whether to skip test
        if self._skip_test(unique_id):
            return False

        if function:
            # Return obj will contain any caught exceptions.
            return_obj = self._execute_test_function(function_description, function, function_args, has_return,
                                                     unique_id, stop_timeout)

            # TODO : Add any caught exception as 'error' log for QCS

        if debug:
            self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "Debug", debug,
                                                                  os.path.basename(__file__) + " - " +
                                                                  sys._getframe().f_code.co_name, uId=""))
        if warning:
            self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "warning", warning,
                                                                  os.path.basename(__file__) + " - " +
                                                                  sys._getframe().f_code.co_name, uId=""))

        if test_description:
            self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "testDescription",
                                                                  test_description, os.path.basename(__file__)
                                                                  + " - " + sys._getframe().f_code.co_name,
                                                                  uId=unique_id))

        # return object if there is one, else return True.
        if return_obj.return_item is None:
            return True

        return return_obj.return_item

    def check_point(self, unique_id, description, function, function_args, has_return=False):

        if self.execution_mode == "seek":
            self.number_of_test_points += 2
            return

        if not dtsGlobals.continueTest:
            return

        # checking whether to skip test
        if self._skip_test(unique_id):
            return

        return_obj = ReturnObject()

        if not self.document_mode:
            return_obj.return_item = function(**function_args)
        else:
            return_obj.return_item = True

        self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "testResult", "description : " + description,
                                                              os.path.basename(
                                                                  __file__) + " - " + sys._getframe().f_code.co_name,
                                                              test_result=str(return_obj.return_item),
                                                              uId=unique_id))

        if has_return:
            return return_obj.return_item

    def _add_quarch_command(self, command, quarch_device, expected_response="OK"):
        result = quarch_device.sendCommand(command)

        self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "quarchCommand",
                                                              "Quarch Command: " + command + " - Response: " + result,
                                                              os.path.basename(
                                                                  __file__) + " - " + sys._getframe().f_code.co_name,
                                                              {"debugLevel": 1,
                                                               "textDetails": "Executing command on module"},
                                                              uId=""))
        # Verify that the command executed as expected
        if result != expected_response:
            self.comms.sendMsgToGUI(
                self.comms.create_request_log(time.time(), "error", "Error executing Torridon command",
                                              os.path.basename(__file__) + " - " + sys._getframe().f_code.co_name,
                                              {"debugLevel": 2, "response_type": str(type(result)),
                                               "response": result,
                                               "command": command}, uId=""))
            result = False

        return result

    def _execute_test_function(self, function_description, function, function_args, has_return, unique_id,
                               stop_timeout):

        return_value = ReturnObject()

        if function_description:
            self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "testDescription",
                                                                  str(function_description),
                                                                  "{} + {}".format(os.path.basename(__file__),
                                                                                   sys._getframe().f_code.co_name),
                                                                  uId=unique_id))

        # Will need to check - if it has a return i may need said results
        if not self.document_mode:
            if stop_timeout:
                self.comms.send_stop_timeout()
            try:
                return_value.return_item = function(**function_args)
            except Exception as e:
                return_value.error = e
                traceback.print_exc()
                print(e)
                self.comms.create_request_log(time.time(), "error", "Error executing Function",
                                              os.path.basename(__file__) + " - " + sys._getframe().f_code.co_name,
                                              {"debugLevel": 2, "response_type": str(type(e)),
                                               "exception": str(e),
                                               "function": str(function)}, uId="")
                logging.warning("Exception caught during execution of function : " + str(function))
                return_value.return_item = None
            if stop_timeout:
                self.comms.send_start_timeout()

        if has_return:
            # All items that return dictionary as
            return_desc = None
            if self.document_mode:
                return_desc = "Document Mode Placeholder value"
            elif isinstance(return_value.return_item, dict):
                if "key_return" in return_value.return_item.keys():
                    return_desc = return_value.return_item["key_return"]
                    return_value.return_item = return_value.return_item["value_return"]
                # else:
                #     return_desc = str(return_value.return_item)

            if return_desc is not None:
                self.comms.sendMsgToGUI(self.comms.create_request_log(time.time(), "TestReturn",
                                                                      "Return value of Function : " + return_desc,
                                                                      os.path.basename(
                                                                          __file__) + " - " + sys._getframe().f_code.co_name,
                                                                      uId=unique_id))

        return return_value

    def test_check_link_speed(self, drive, quarch_module, link_speed):
        """
        Checks if the pcie device has same lane width and link speed
        """
        # Moving this so it's not checked on every search for device
        if isinstance(drive, DriveWrapper):
            if "PCIE" in drive.drive_type:
                # default value is none. If no custom value assigned, verify_drive_link will handle value assignment
                return self.my_host_info.verify_drive_link(quarch_module, expected_speed=link_speed)
            else:
                self.test_point(debug="SAS drive does not support Lane Width and Link Speed detection")
                return True

    def test_check_lane_width(self, drive, quarch_module, lane_width):
        """
        Checks if the pcie device has same lane width and link speed
        """
        # Moving this so it's not checked on every search for device
        if isinstance(drive, DriveWrapper):
            if "PCIE" in drive.drive_type:
                # default value is none. If no custom value assigned, verify_drive_link will handle value assignment
                return self.my_host_info.verify_drive_lane_width(quarch_module, expected_width=lane_width)
            else:
                self.test_point(debug="SAS drive does not support Lane Width and Link Speed detection")
                return True

    def test_wait_for_enumeration(self, enumeration, drive, ontime=0, offtime=0):

        if enumeration:
            enum_time = checkDriveState(drive, True, ontime)
        else:
            enum_time = checkDriveState(drive, False, offtime)

        return enum_time

    def request_qps(self):
        if self.document_mode:
            return True
        if not isQpsRunning(host=dtsGlobals.GUI_TCP_IP, timeout=5):
            path = inspect.getfile(inspect.currentframe())
            print("starting qps")
            self.comms.sendMsgToGUI(self.comms.create_request_function("start_qps", path))
            if not isQpsRunning(host=dtsGlobals.GUI_TCP_IP, timeout=25):
                self.comms.send_stop_test(reason="No running QPS instance found")
                print("Unable to continue test, no running instance of QPS found")
                return False
        return True

    def select_drive(self, drive_type="all"):
        """
        :param drive_type: Accepted values : "sas", "pcie", "all"
        :return:
        """

        self.test_point(self.test_id.gen_next_id(), function_description="Finding available drives",
                        function=self.my_host_info.get_drives, has_return=True, function_args={'drive_type': drive_type})
        user_choice = self.test_point(self.test_id.gen_next_id(), has_return=True,
                                      function_description="User selection – Drive to use",
                                      function=self.select_item,
                                      function_args={"description": "",
                                                     "window_mode": "drive",
                                                     "drive_dictionary": self.my_host_info.device_list})

        if "rescan" in str(user_choice):
            return self.select_drive()
        if "abort" in str(user_choice):
            return None

        drive = self.test_point(self.test_id.gen_next_id(), has_return=True,
                                function_description="User selection – Drive to use",
                                function=self.my_host_info.get_drive_from_choice,
                                function_args={"selection": user_choice,
                                               "report_dict": self.report})

        if isinstance(drive, DriveWrapper):
            self.comms.sendMsgToGUI(
                self.comms.create_request_log(time.time(), "testReturn",
                                              "User Select drive : {0} , {1}".format(drive.identifier_str,
                                                                                     drive.description),
                                              os.path.basename(__file__) + " - " + sys._getframe().f_code.co_name,
                                              uId=self.test_id.gen_next_id()))

        if self.document_mode:
            drive = DriveWrapper()

        if drive.drive_type in "Unknown":
            self.test_point(warning="Unknown drive types do not support Lane Width and Link Speed detection")
        elif drive.drive_type in "sas":
            self.test_point(warning="QCS does not currently support Lane Width and Link Speed detection for Sas drives")
        else:
            self.test_point(self.test_id.gen_next_id(), has_return=True,
                            function_description="User selection – Quarch module to use",
                            function=self.my_host_info.store_initial_drive_stats,
                            function_args={"drive": drive})

        return drive

    def select_quarch_module(self, use_qps=False, ip_lookup=None, filter_module=None):
        """
        :param use_qps: whether to use python or qps
        :param ip_lookup: IP to send discovery packet to - utilized by running test only.
        :param filter_module:
        :return:
        """
        function = get_quarch_modules_qps if use_qps else scanDevices

        function_args_dict = _create_scan_dict(use_qps, ip_lookup, filter_module)

        software_type = "qps" if use_qps else "py"

        module_list = self.test_point(self.test_id.gen_next_id(), has_return=True,
                                      function_description="User selection – Quarch breaker module to use",
                                      function=function, function_args=function_args_dict)

        user_choice = self.test_point(self.test_id.gen_next_id(), has_return=True,
                                      function_description="User selection – Quarch module to use",
                                      function=self.select_item,
                                      function_args={"description": "",
                                                     "window_mode": software_type,
                                                     "module_dictionary": module_list})
        # check for rescan / rescan with ip lookup
        if "abort" in str(user_choice):
            return None
        if "rescan" in str(user_choice):
            if "==" in str(user_choice):
                ip_lookup = user_choice[user_choice.index("==") + 2:]
            return self.select_quarch_module(use_qps=use_qps, ip_lookup=ip_lookup, filter_module=filter_module)

        return_val = self.test_point(self.test_id.gen_next_id(), has_return=True, function=get_module_from_choice,
                                     function_description="User selection – Quarch breaker module to use",
                                     function_args={"connection": user_choice, "powerOnDevice": True, "is_qps": use_qps,
                                                    "report_dict": self.report, "return_val": True})

        self.comms.sendMsgToGUI(
            self.comms.create_request_log(time.time(), "testReturn",
                                          "User Select module : {0}".format(user_choice),
                                          os.path.basename(__file__) + " - " + sys._getframe().f_code.co_name,
                                          uId=self.test_id.gen_next_id()))

        if self.document_mode:
            return_val = "placeholder"

        return return_val

    def select_item(self, description, window_mode, drive_dictionary=None, module_dictionary=None):
        """
        :param description: Description of selection
        :param window_mode: qps / py / drive
        :param module_dictionary:
        :param drive_dictionary:
        :return: Selected item from the user
        """

        if window_mode == "qps":
            formatted_qps_modules_dict = {}
            for quarch_module in module_dictionary:
                if "rest" in quarch_module:
                    continue
                connection_without_interface = quarch_module[quarch_module.rindex(":") + 1:]
                formatted_qps_modules_dict[quarch_module.replace("::", ":")] = connection_without_interface
            module_dictionary = formatted_qps_modules_dict

        self.comms.sendMsgToGUI(self.comms.create_request_gui(title="user selection", description=description,
                                                              window_type="SelectionGrid", window_mode=window_mode,
                                                              dict_of_drives=drive_dictionary,
                                                              dict_of_modules=module_dictionary))

        while dtsGlobals.choiceResponse is None and dtsGlobals.continueTest is True:
            time.sleep(0.25)

        choice = str(dtsGlobals.choiceResponse)
        selection = choice.split("::")
        selection = selection[1]
        if window_mode == "qps":
            selection = selection.replace(":", "::")

        return selection


class ReturnObject:
    def __init__(self):
        self.return_item = None
        self.error = None
        self.user_choice = None
